# -*- coding: utf-8 -*-
#----------------------------------------------------------------------
#     This file is a part of OpenSynergy System
#
#     F I L E  N A M E:
#        KlienciModel.py
#     F U N C T I O N S:
#        This is the Model module for Contacts component
#     A U T H O R:
#        Sebastian Żurek, 02-02-2008
#     M O D I F I E D  B Y:
#        (add youtself here if you are modifing this file)
#
#     L I C E N S E
#
#     Copyright (C) 2007 Sebastian Żurek
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#----------------------------------------------------------------------

u"""

MVC scheme Model element to access Users data required to Login.

Class LoginModel:\n\t Provides public methods to access Users data

"""

import wx.lib.pubsub as pubsub
import types
import string
from Synergy import Model
from Synergy import Exceptions
from psycopg2.extensions import adapt, register_adapter, AsIs


class KlienciModel(Model.Model):

    def __init__(self,*args,**kwargs):
        
        """Creates Model object instance.

        """
        super(KlienciModel,self).__init__(*args,**kwargs)

                      

# - Table Creation Fields ----------------------------------------------------------

    def GetTables(self):
        """ Returns list with tables names to be created during system initialization

        """
        return ['contacts', 'contacts_groups']


    def _createTable_contacts_postgres(self):
        """ Returns SQL query to create clients table with PostgreSQL database engine

        """

        # WARNING! Tha name of table clients has been changed to contacts!
        # I've added some new fields and the meaning of some fields were changed:
        #  - parent: the integer number that may be a cid of a parent or -1 if
        #            it is root contact
        #  - website: the website of a contacts
        #  - group: an integer number - 0 for ordinary (non-client) contacts, 1 for physical person that
        #           is client and 2 for economic cient.
        #
        # tHE IDEAS
        # - private and public contacts
        # - skype, gg,...
        return """CREATE TABLE contacts(
                        cid SERIAL,            
                        full_name VARCHAR NOT NULL,
                        identifier VARCHAR(15) NOT NULL,
                        nip  VARCHAR(10),
                        regon VARCHAR(14), 
                        bank_name VARCHAR,
                        bank_account VARCHAR,
                        city VARCHAR,
                        zip VARCHAR,
                        post VARCHAR,
                        address VARCHAR,
                        email VARCHAR,
                        website VARCHAR,
                        tel VARCHAR,
                        grouping INTEGER DEFAULT 0,
                        parent INTEGER DEFAULT -1,
                        CONSTRAINT contacts_pkey PRIMARY KEY (cid),
                        CONSTRAINT uinque_contact_identifier UNIQUE (identifier)
                        ) WITH OIDS"""


    def _createTable_contacts_groups_postgres(self):
        """ Returns SQL query to create contacts groups table with PostgreSQL database engine

        """
        return """CREATE TABLE contacts_groups(
                        cgid SERIAL,            
                        full_name VARCHAR NOT NULL,                        
                        identifier VARCHAR(15) NOT NULL,
                        description TEXT,
                        CONSTRAINT cg_pkey PRIMARY KEY (cgid),
                        CONSTRAINT uinque_cg_identifier UNIQUE (identifier)
                        ) WITH OIDS"""


# - Getter Methods  --------------------------------------------------------------

    def GetContacts(self, parent = -1, search = '', group = None, children = True, basic=True, address=True, contact=True, invoice=True):
        """ Returns a dictionary with Contacts info.

        Keyword arguments:
        - key: a main key used to select contacts. It may be a list (of integers or strings) for multiple clients selection or a
               single integer or string representing a certain client.
        - keyType: represents a key to be used for a selection: one cane select using the cid's or using the contacts unique identifiers
        - search: a string element used to search for,
        - children: if this bool option is set (default value), the method will return the list of all
                    children of the contact

        """
        query,colnames = self._RenderGetContactsSelect(basic,address,contact,invoice)
        condition, values = self._RenderGetContactsConditions(search = search,group = group, parent = parent)
        query += condition

        self._cursor.execute(query,values)

        contacts_tuples = self._cursor.fetchall()
        contacts_list = []
        
        if contacts_tuples:
            for contact in contacts_tuples:
                contact_dict = {}
                for i,colname in enumerate(colnames):
                    # We map all string to unicode objects...
                    contact_dict[colname] = contact[i] if type(contact[i])!=types.StringType else contact[i].decode('utf-8')

                # Now, we will query for children of a contact... and we'll let the recursion to do all the hard work
                child_list = []
                if children:
                    child_list = self.GetContacts(contact_dict['cid'] , search ,group,children ,basic,address,contact,invoice)
                
                # The contacts_dict is extended by the new Contact object
                contacts_list.append(Contact(data = contact_dict,children = child_list))

        return contacts_list


    def GetContact(self,key = None, keyType='cid', children = False, basic=True,address=True,contact=True,invoice=True):
        """ Returns a dictionary with Contact info.

        Keyword arguments:
        - key: a main key used to select contacts. It may be a list (of integers or strings) for multiple clients selection or a
               single integer or string representing a certain client.
        - keyType: represents a key to be used for a selection: one cane select using the cid's or using the contacts unique identifiers
        - children: if this bool option is set (default value), the method will return the list of all
                    children of the contact

        """
 
        query,colnames = self._RenderGetContactsSelect(basic,address,contact,invoice)
        condition, values = self._RenderGetContactsConditions(key,keyType)
        query += condition

        self._cursor.execute(query,values)

        contact = self._cursor.fetchone()
        if contact:
            contact_dict = {}
            for i,colname in enumerate(colnames):
                # We map all string to unicode objects...
                contact_dict[colname] = contact[i] if type(contact[i])!=types.StringType else contact[i].decode('utf-8')

                # Now, we will query for children of a contact... and we'll let the recursion to do all the hard work
                child_list = []
                if children:
                    child_list = self.GetContacts(parent = contact_dict['cid'] , children = children,basic = basic,address = address,
                                                                contact = contact, invoice = invoice)
                
                # Contact object is returned
            return  Contact(data = contact_dict, children = child_list)
        else:
            return None


    def _RenderGetContactsSelect(self,basic,address,contact,invoice):
        arguments = locals()
        arguments.pop('self')
        
        columns = {'basic': ('cid','grouping','identifier', 'full_name'),
                   'address':('city', 'zip', 'post', 'address'),
                   'contact':('email', 'tel', 'website'),
                   'invoice': ('nip', 'regon', 'bank_name', 'bank_account')}
        selection_list = []
        for column in columns:
            if arguments[column]:
                selection_list.extend(columns[column])

        return 'SELECT '+string.join(selection_list,', ') + ' FROM contacts ',selection_list

    def _RenderGetContactsConditions(self,key = None, keyType = 'cid',search = '',group = None, parent = -1):
        """Returns rendered SQL subelement of SELECT query responsible for
        conditional statement.

        Keyword arguments:
        - cid: a number or a list of numbers (for mulitple selection) representing clients
               primary key
        - identifier: a string or list of string (for multiple selection) representing clients
                      short names (identifiers)
        
        """

        # This is used for the determination of data type for cid and identifier
        # if cid or identifier is a list, whe will use 'IN' operator in selection
        # conditional query, if it is just a single number or a string, we will ues '=' operator
        key_operator = 'IN' if type(key)==types.ListType else '='

        # Simmilar situation is for groups. Once can specife either the list of groups or the
        # single number
        group_operator = 'IN' if type(group)==types.ListType else '='

        
        # Query conditions -----------------------------------------------
        condintion_to_return={'string':[],
                              'values':[]
                              }
        # Keys in this dictionary have the names after the method arguments names
        conditions=({'on':bool(key),
                     'string':"%s %s %%s"%(keyType,key_operator),
                     'values':(key,)
                     },                                          
                    {'on':bool(search),
                     'string':'identifier ~* %s OR full_name ~* %s',
                     'values': (search,search)
                     },
                    {'on':bool(group),
                     'string':'grouping %s %%s'%(group_operator,),
                     'values': (group,)
                     },
                    {'on':True,
                     'string':'parent = %s',
                     'values': (parent,)
                     }
                    )
        
        for cond in conditions:
            if cond['on']:
                condintion_to_return['string'].append(cond['string'])
                condintion_to_return['values'].extend(cond['values'])
        if condintion_to_return['string']:
            condintion_to_return['string']='WHERE '+string.join(condintion_to_return['string'],' AND ')
        else:
            # If thre are no atom condition in the conditions list, I return
            # an empty string...
            condintion_to_return['string'] = ''

        return condintion_to_return['string'],condintion_to_return['values']



    def GetClientsWithType(self,type):
        """ Returns Clients data in dictionary

        """

        query="select  full_name, identifier, nip, regon, bank_name, bank_account, city, zip, post, address, email, tel, cid from clients WHERE type=%s"
        self._cursor.execute(query,(type,))
        result=self._cursor.fetchall()
        to_return=[]
        
        if self._dbType=='postgres':
            if result:
                for row in result:
                    r_dict={            
                        'full_name':row[0].decode('utf-8'),
                        'identifier':row[1].decode('utf-8'),
                        'nip':row[2],
                        'regon':row[3],
                        'bank_name':row[4].decode('utf-8'),
                        'bank_account':row[5],
                        'city':row[6].decode('utf-8'),
                        'zip':row[7],
                        'post':row[8].decode('utf-8'),
                        'address':row[9].decode('utf-8'),            
                        'email':row[10],
                        'tel':row[11],
                        'cid':row[12]
                        }
                    to_return.append(r_dict)                
        return to_return

    def GetClientIdentifier(self,cid):
        """ Returns Clients data in dictionary

        """
        query="SELECT identifier FROM clients WHERE cid=%s"
        self._cursor.execute(query,(cid,))
        return self._cursor.fetchone()[0].decode('utf-8')


    def GetClientsIdentifiersList(self):
        """ Returns Clients data in dictionary

        """
        query="SELECT identifier FROM clients ORDER BY identifier"
        self._cursor.execute(query)
        zwrot=[]
        result=list(self._cursor.fetchall())
        if self._dbType=='postgres':
            if result:
                for i in range(len(result)):
                    zwrot.append(result[i][0].decode('utf-8'))

        return zwrot

    def GetClientType(self,cid):
        """ Returns Clients data in dictionary

        """
        query="SELECT type FROM clients WHERE cid=%s"
        self._cursor.execute(query,(cid,))
        return self._cursor.fetchone()[0]


    def GetClientIdByIdentifier(self,identifier):
        query="SELECT cid FROM clients WHERE identifier=%s"
        self._cursor.execute(query,(identifier,))
        result=self._cursor.fetchone()        
        if result:
            return result[0]
        else:
            return None

    def GetClientByIdentifier(self,identifier):
        """ Returns Clients data in dictionary

        """

        query="select  full_name,  nip, regon, bank_name, bank_account, city, zip, post, address, email, tel, cid, type FROM clients WHERE identifier=%s"
        self._cursor.execute(query,(identifier,))
        result=self._cursor.fetchone()        
        if self._dbType=='postgres':
            if result:
                r_dict={            
                    'full_name':result[0].decode('utf-8'),
                    'nip':result[1],
                    'regon':result[2],
                    'bank_name':result[3].decode('utf-8'),
                    'bank_account':result[4],
                    'city':result[5].decode('utf-8'),
                    'zip':result[6],
                    'post':result[7].decode('utf-8'),
                    'street':result[8].decode('utf-8'),            
                    'email':result[9],
                    'tel':result[10],
                    'cid':result[11],
                    'type':result[12]
                    }
                return r_dict


#- Insert Methods  --------------------------------------------------------------

    def InsertContact(self,contact):
        """ Insert Contact data into table.

        Keywords arguments:

        contact - Contact class object
        
        """
        data = contact.GetRawData()
        # I remove cid because it's None
        data.pop('cid')
        
        try:
            self._cursor.execute(self.ParseInsertQuery(data,'contacts'),data.values())
            self._db.commit()         
            pubsub.Publisher().sendMessage("Contacts.Model.Events", data=('NewContact',[contact]))        
        except Exception, error:            
            self._db.rollback()            
            raise Exceptions.DatabaseError(error.message)


    def UpdateContact(self,data,where):
        """Update Contact data

        Arguments:

        data - dictionary where keys are table columns
        where - dictionary holding where statement
        
        """

        try:
            values=data.values()
            values.append(where.values()[0])
            self._cursor.execute(self.ParseUpdateQuery(data,'clients',where),values)
            self._db.commit()
            pubsub.Publisher().sendMessage("Contacts.Model.Events", data=('ContactUpdated',[]))        
        except Exception, error:            
            self._db.rollback()            
            raise Exceptions.DatabaseError(error.message)


    def UpdateUserAccessById(self,data):
        self._models['Users'].UpdateUserAccessById(data)

#- Delete MEthods  --------------------------------------------------------------

    def DeleteContact(self,contact):
        query="DELETE FROM contacts WHERE cid = %s"

        try:
            self._cursor.execute(query,(contact.GetRawData()['cid'],))
            self._db.commit()
            pubsub.Publisher().sendMessage("Contacts.Model.Events", data=('ContactRemoved',[contact]))        
        except Exception, error:            
            self._db.rollback()            
            raise Exceptions.DatabaseError(error.message)

#- Special Methods  --------------------------------------------------------------


class Contact(object):
    """Contact object class.

    """
    _types_dict = {0: u'Zwykły kontakt',
                   1: u'Klient Gospodarczy',
                   2: u'Klient Fizyczny'}

    _model = KlienciModel()
    
    def __init__(self, cid = None, data = None, children = []):
        """Init the contact object.

        Keyword arguments:
        cid - and ContactID number
        data - data dictionary that may be used to initialize the contact
        
        """

        # All contact data is saved in dictionary.
        self._data = {
                'cid': None,            
                'full_name': None,
                'identifier': None,
                'nip': None,
                'regon':None, 
                'bank_name':None,
                'bank_account':None,
                'city':None,
                'zip':None,
                'post':None,
                'address':None,
                'email': None,
                'website':None,
                'tel':None,
                'grouping':0,
                'parent':-1
                }
        
        self._children = children
        
        # If optional argument uid is set, the __init__ method will
        # populate the data dictionary with apropriate data
        if cid != None:
            self._data = self.GetContacts(key = cid, keyType = 'cid').GetRawData()
        elif data != None:
            for key in self._data.keys():
                if data.has_key(key):
                    self._data[key] = data[key]

    def GetData(self):
        """Return the translated data Dictionary.

        The translation means, that all symbolic data values (such as type which is
        represented in db by INTEGER data type) is translated to a meaningfull string.

        """
        translatedDict = {}
        for field in self._data.keys():
            if field == 'grouping':
                translatedDict[field] = self._types_dict[self._data[field]]
            else:
                translatedDict[field] = self._data[field]
        return translatedDict

    def SetData(self,data):
        """Sets the Contact data dictionary

        """
        self._data = data

    def GetRawData(self):
        """Return raw data dictionary - each key is the column name in table
        'clients', and each related value represents the value in the db table

        """
        return self._data


    def SetParent(self,parent):        
        self._data['parent'] = parent.GetRawData()['cid']
        print 'Ustawiłem parenta :',parent


    def GetChildCount(self):
        return len(self._children)

    def GetChildren(self):
        return self._children


    def GetIdentifier(self):
        return self._data['identifier']

    def GetFullName(self):
        return self._data['full_name']

    def Save(self):
        """Save the Client in database.

        One can simply store the client data (self._data) in database
        by calling this method. If the client does not exist self._data['uid']=None
        it will be inserted, otherwise its data will be updated.
        
        """

        if self._data['cid'] !=None:
            self._model.UpdateContact(self)
        elif self._data['cid'] == None:
            self._model.InsertContact(self)

    def Remove(self):
        """Deletes the Client from the database.

        """
        if self._data['cid'] !=None:
            self._model.DeleteContact(self)
        else:
            # TO DO: add some try-except conditions and
            # eventually raise an error
            #
            #
            print 'Nie mam co usuwac'
